/*
 * FreeRTOS Bananapi M3 port
 * (c) 2020 Dr.-Ing. Ingo D. Rullhusen <d01c@freeshell.de>
 *
 * BPI M3 port is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BPI M3 port is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FreeRTOS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "FreeRTOS.h"
#include "task.h"
#include "partest.h"

#include "partest_ext.h"

#define partstNUM_LEDS ( 2 ) // true LEDs

/*-----------------------------------------------------------
 * Simple "parallel port" IO routines: LEDs on PMU not GPIO.
 * uxLED: 0 = green, 1 = blue, 2 = PC7, 3 = PC4
 *-----------------------------------------------------------*/

static int rsb_await_trans(void)
{
   unsigned long tmo = xTaskGetTickCount() + 1000;
   uint32_t stat;
   int ret;

   while (1)
   {
      stat = RSB_STAT;
      if ( stat & 0x4 ) // LOAD_BSY
      {
         ret = -3;
         break;
      }
      if ( stat & 0x2 ) // TRANS_ERR)
      {
         ret = -2;
         break;
      }
      if ( stat & 0x1 ) // TRANS_OVER
      {
         ret = 0;
         break;
      }
      if ( xTaskGetTickCount() > tmo )
      {
         ret = -1;
         break;
      }
   }
   RSB_STAT = stat; /* Clear status bits */

   return(ret);
}

void vParTestInitialise( void )
{

   portENTER_CRITICAL()
   RSB_SADDR = 0x002d03a3;
   RSB_CMD = 0xe8; // SRTA
   RSB_CTRL = 0x00000080;
   portEXIT_CRITICAL()

   rsb_await_trans();

   portENTER_CRITICAL()
   RSB_DADDR0 = 0x90; //  green
   RSB_DLEN = 0x00000000; // write/0+1
   RSB_DATA0 = 0x00;
   RSB_CMD = 0x4e; // WR8
   RSB_CTRL = 0x00000080; // START_TRANS
   portEXIT_CRITICAL()

   rsb_await_trans();

   portENTER_CRITICAL()
   RSB_DADDR0 = 0x92; //  blue
   RSB_DLEN = 0x00000000; // write/0+1
   RSB_DATA0 = 0x00;
   RSB_CMD = 0x4e; // WR8
   RSB_CTRL = 0x00000080; // START_TRANS
   portEXIT_CRITICAL()

   PC_CFG0_REG = 0x17717777;

}


void vParTestSetLED( unsigned portBASE_TYPE uxLED, signed portBASE_TYPE xValue )
{

   switch (uxLED)
   {

      case 0:
      {
         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DADDR0 = 0x90; // green
         RSB_DLEN = 0x00000000; // write/0+1
         RSB_DATA0 = ( xValue ? 0x01 : 0x00 );
         RSB_CMD = 0x4e; // WR8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         break;
      }

      case 1:
      {
         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DADDR0 = 0x92; // blue
         RSB_DLEN = 0x00000000; // write/0+1
         RSB_DATA0 = ( xValue ? 0x01 : 0x00 );
         RSB_CMD = 0x4e; // WR8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         break;
      }

      case 2:
      {
         if ( xValue ) { GPIO_IOSET |= 0x00000080; }
         else          { GPIO_IOSET &= 0xffffff7f; }
         break;
      }

      case 3:
      {
         if ( xValue ) { GPIO_IOSET |= 0x00000010; }
         else          { GPIO_IOSET &= 0xffffffef; }
         break;
      }

   }

}


void vParTestToggleLED( unsigned portBASE_TYPE uxLED )
{

   switch (uxLED)
   {

      case 0:
      {
         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DADDR0 = 0x90; // green
         RSB_DLEN = 0x00000010; // read/0+1
         RSB_CMD = 0x8b; // RD8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DLEN = 0x00000000; // write/0+1
         RSB_DATA0 = ( RSB_DATA0&0x000000ff ? 0x00 : 0x01 );
         RSB_CMD = 0x4e; // WR8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         break;
      }

      case 1:
      {
         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DADDR0 = 0x92; // blue
         RSB_DLEN = 0x00000010; // read/0+1
         RSB_CMD = 0x8b; // RD8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         rsb_await_trans();

         portENTER_CRITICAL()
         RSB_DLEN = 0x00000000; // write/0+1
         RSB_DATA0 = ( RSB_DATA0&0x000000ff ? 0x00 : 0x01 );
         RSB_CMD = 0x4e; // WR8
         RSB_CTRL = 0x00000080; // START_TRANS
         portEXIT_CRITICAL()

         break;
      }

      case 2:
      {
         GPIO_IOSET ^= 0x00000080;
         break;
      }

      case 3:
      {
         GPIO_IOSET ^= 0x00000010;
         break;
      }

   }

}


void vPowerOff( void )
{
   rsb_await_trans();

   portENTER_CRITICAL()
   RSB_DADDR0 = 0x32;
   RSB_DLEN = 0x00000000; // write/0+1
   RSB_DATA0 = 0x80;
   RSB_CMD = 0x4e; // WR8
   RSB_CTRL = 0x00000080; // START_TRANS
   portEXIT_CRITICAL()

}


