/*
 * FreeRTOS Bananapi M3 port
 * (c) 2020 Dr.-Ing. Ingo D. Rullhusen <d01c@freeshell.de>
 *
 * BPI M3 port is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BPI M3 port is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FreeRTOS; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <string.h>

#include <sys/unistd.h>

#include "blink_task.h"
#include "cmd_task.h"
#include "mon_task.h"

#include "serial.h"
#include "partest_ext.h"


extern MonTaskParameter xMon[8];


#define OS_PROMPT "FreeRTOS> "

#define uart_timeout 1000
#define cmd_buff_size 256

static int cmd_buff_n = 0;
static char cmd_buffer[ cmd_buff_size ];


static void cmd_help( void )
{
   printf( "\nhelp\n" );
   printf( "   show this help\n" );
   printf( "\ninterval [green|blue|pc7|pc4]\n" );
   printf( "   show blink interval in ticks\n");
   printf( "\ninterval [green|blue|pc7|pc4] <ticks>\n" );
   printf( "   set blink interval to <ticks>\n");
   printf( "\npoweroff\n" );
   printf( "   switch off system immediately\n" );
   printf( "\nreboot\n" );
   printf( "   reboot system immediately\n" );
   printf( "\nsgi <CPUID> <INTID>\n" );
   printf( "   send software generated interrupt <INTID 0...15> to CPU <CPUID 0...7>\n" );
   printf( "\nstatus\n" );
   printf( "   show status information\n" );
   printf( "   WARNING: Scheduler remaining suspended for an extended period!\n" );
   printf( "\nversion\n" );
   printf( "   show version information\n" );
   printf( "\n" );
}

static void cmd_version( void )
{
   printf( "\n" );
   printf( "FreeRTOS: "tskKERNEL_VERSION_NUMBER"\n" );
   printf( "App     : "App_Version"\n" );
   printf( "\n" );
}

extern BlinkTaskParameter xGreen;
extern BlinkTaskParameter xBlue;
extern BlinkTaskParameter xPc7;
extern BlinkTaskParameter xPc4;

static void cmd_set_interval( BlinkTaskParameter *pxParameter,
                              TickType_t xInterval )
{
   pxParameter->interval = xInterval;
}

static void cmd_get_interval( BlinkTaskParameter *pxParameter )
{
   TickType_t interval_t;

   interval_t = pxParameter->interval;

   printf( "%u\n", interval_t );
   printf( "\n" );
}

static void cmd_status( void )
{
   printf( "\n" );
   printf( "WARNING: Scheduler remaining suspended for an extended period!\n" );
   printf( "\n" );

   xMon[0].ctrl = 1;
   while ( xMon[0].ctrl != 2 ) { taskYIELD(); }
   printf( "%s", xMon[0].answer );
   xMon[0].ctrl = 0;

   printf( "\n" );

   for ( int c=1; c<8; ++c )
   {
      xMon[c].ctrl = 1;

      TickType_t start = xTaskGetTickCount();
      int ctrl;
      while ( (ctrl = xMon[c].ctrl) != 2 )
      {
         if ( ctrl != 1 )
         {
            snprintf( xMon[c].answer, MonResponseBufferSize, "<not ready>\n" );
            break;
         }
         taskYIELD();
         if ( xTaskGetTickCount() - start > pdMS_TO_TICKS(3000) )
         {
            snprintf( xMon[c].answer, MonResponseBufferSize, "<timeout>\n" );
            break;
         }
      }
      printf( "%s", xMon[c].answer );

      xMon[c].ctrl = 0;

      printf( "\n" );
   }

}


static void command_interpreter( const char *cmd, int size )
{
   (void) size;
   printf("command:%s\n",cmd);

   char *token1 = strtok( (char *)cmd, " \n\r\t\0" );
   if ( ! token1 ) return;

   if      ( strcasecmp( token1, "help"    ) == 0 )
   { cmd_help(); return; }
   else if ( strcasecmp( token1, "version" ) == 0 )
   { cmd_version(); return; }
   else if ( strcasecmp( token1, "interval" ) == 0 )
   {
      char *token2 = strtok( NULL, " \n\r\t\0" );

      char *token3 = strtok( NULL, " \n\r\t\0" );
      int interval, items = 0;
      if ( token3 ) { items = sscanf( token3, "%u", &interval ); }

      if      ( strcasecmp( token2, "green" ) == 0 )
      {
         if ( items ) { cmd_set_interval( &xGreen, interval ); }
         else         { cmd_get_interval( &xGreen );           }
         return;
      }      
      else if ( strcasecmp( token2, "blue"  ) == 0 )
      {
         if ( items ) { cmd_set_interval( &xBlue, interval ); }
         else         { cmd_get_interval( &xBlue );           }
         return;
      }
      else if ( strcasecmp( token2, "pc7"   ) == 0 )
      {
         if ( items ) { cmd_set_interval( &xPc7, interval ); }
         else         { cmd_get_interval( &xPc7 );           }
         return;
      }      
      else if ( strcasecmp( token2, "pc4"   ) == 0 )
      {
         if ( items ) { cmd_set_interval( &xPc4, interval ); }
         else         { cmd_get_interval( &xPc4 );           }
         return;
      }
   }
   else if ( strcasecmp( token1, "sgi" ) == 0 )
   {
      char *token2 = strtok( NULL, " \n\r\t\0" );
      char *token3 = strtok( NULL, " \n\r\t\0" );

      int cpuid = 0, intid = 0, items = 0;
      if ( token2 ) { items = sscanf( token2, "%u", &cpuid ); }
      if ( token3 ) { items = sscanf( token3, "%u", &intid ); }

      if ( items ) { GICD_SGIR = ( 0x0 | (0x1<<(16+cpuid)) | (0xf & intid) ); }

      return;
   }
   else if ( strcasecmp( token1, "status" ) == 0 )
   { cmd_status(); return; }
   else if ( strcasecmp( token1, "reboot" ) == 0 )
   {
      printf( "Reboot...\n" );
      fflush( NULL );
      vTaskDelay( pdMS_TO_TICKS(50) );

      portENTER_CRITICAL();
/* watchdog reset */
      RWDOG_CFG_REG = 0x1; // to whole system
      RWDOG_MODE_REG = 0x1; // enable 0.5s
      RWDOG_CTRL_REG = (0x0a57 << 1) | 0x1; // key + start
      while(1) {}
      portEXIT_CRITICAL();

      return;
   }
   else if ( strcasecmp( token1, "poweroff" ) == 0 )
   {
      printf( "Switch off.\n" );
      fflush( NULL );
      vTaskDelay( pdMS_TO_TICKS(50) );

      vPowerOff();

      return;
   }
   else if ( strcasecmp( token1, "xxx" ) == 0 )
   {

      for ( int i=0; i<32; ++i )
      {
         printf("release spinlock %i\n", i );
         SPINLOCK_LOCK_REG(i) = 0;
      }

      return;
   }

   printf( "Unknown command.\n" );

}


portTASK_FUNCTION( CmdTaskFunction, pvParameters )
{
   CmdTaskParameter *param = (CmdTaskParameter*)pvParameters;

   param->lastStart = xTaskGetTickCount();

   printf( OS_PROMPT );
   fflush( NULL );

   for( ; ; )
   {
      signed char cRxedChar;
      while ( xSerialGetChar( 0, &cRxedChar, uart_timeout ) == pdTRUE )
      {
// echo handling
         if ( cRxedChar == 0x1b /* ESC */ )
         {
            continue;
         }
         else if ( cRxedChar == '\b' /* BS */ || cRxedChar == 0x7f /* DEL */ )
         {
            if ( cmd_buff_n > 0 )
            {
               --cmd_buff_n;
               printf( "\b \b" );
            }
            else
            {
               putchar( '\a' );
            }
         }
         else
         {
            if ( cmd_buff_n >= cmd_buff_size )
            {
               --cmd_buff_n;
               memmove( cmd_buffer, &cmd_buffer[1], cmd_buff_n );
            }

            cmd_buffer[ cmd_buff_n++ ] = cRxedChar;

            putchar( cRxedChar );
         }
         fflush( NULL );

// command handling
         if ( cRxedChar == '\r' )
         {
            cmd_buffer[ cmd_buff_n-1 ] = '\0';
            putchar( '\n' );

            command_interpreter( cmd_buffer, cmd_buff_n );

            printf( OS_PROMPT );
            fflush( NULL );

            cmd_buff_n = 0;
         }

      }

   }

   vTaskDelete( NULL );
}

